--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, tech_ddr_lib, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;

ENTITY io_ddr_driver IS 
  GENERIC (
    g_tech_ddr         : t_c_tech_ddr
   );  
  PORT ( 
    clk                : IN  STD_LOGIC;
    rst                : IN  STD_LOGIC;

    ctlr_init_done     : IN  STD_LOGIC;
    
    ctlr_mosi          : OUT t_tech_ddr_mosi;
    ctlr_miso          : IN  t_tech_ddr_miso;
    
    dvr_en             : IN  STD_LOGIC := '1';
    dvr_wr_not_rd      : IN  STD_LOGIC;       
    dvr_done           : OUT STD_LOGIC; -- Requested wr or rd sequence is done.
    dvr_start_addr     : IN  t_tech_ddr_addr;
    dvr_end_addr       : IN  t_tech_ddr_addr;
   
    wr_fifo_usedw      : IN  STD_LOGIC_VECTOR;
    
    wr_snk_out         : OUT t_dp_siso;
    wr_snk_in          : IN  t_dp_sosi;
    
    rd_src_in          : IN  t_dp_siso;
    rd_src_out         : OUT t_dp_sosi
   );
END io_ddr_driver;


ARCHITECTURE str OF io_ddr_driver IS

  CONSTANT c_address_w          : NATURAL := func_tech_ddr_dq_address_w(g_tech_ddr) + 1;  -- 1 bit added to detect overflow
 
  CONSTANT c_margin             : NATURAL := 2; -- wr_burst_size is updated one cycle after reading actual nof available words.
                                                -- Subtract two (wr_fifo_usedw and wr_burst_size are both registered) so we cannot 
                                                -- post a request for a too large burst size, which could cause the wr_burst state 
                                                -- to be two valid words short.

  SIGNAL req_burst_cycles       : STD_LOGIC_VECTOR(g_tech_ddr.maxburstsize_w-1 DOWNTO 0);
  SIGNAL nxt_req_burst_cycles   : STD_LOGIC_VECTOR(g_tech_ddr.maxburstsize_w-1 DOWNTO 0);

  TYPE t_state_enum IS (s_init, s_idle, s_wait1, s_wait2, s_wait3, s_rd_request, s_wr_request, s_wr_burst);

  SIGNAL state                  : t_state_enum;
  SIGNAL nxt_state              : t_state_enum; 
  SIGNAL prev_state             : t_state_enum; 

  SIGNAL wr_burst_size          : NATURAL;
  SIGNAL rd_burst_size          : NATURAL;

  SIGNAL nxt_wr_burst_size      : NATURAL;
  SIGNAL nxt_rd_burst_size      : NATURAL;

  SIGNAL ctlr_mosi_burstsize    : STD_LOGIC_VECTOR(c_tech_ddr_ctlr_burstsize_w-1 DOWNTO 0);

  SIGNAL i_dvr_done             : STD_LOGIC;
  SIGNAL nxt_dvr_done           : STD_LOGIC;

  SIGNAL start_address          : STD_LOGIC_VECTOR(c_address_w-1 DOWNTO 0);
  SIGNAL end_address            : STD_LOGIC_VECTOR(c_address_w-1 DOWNTO 0);
 
  SIGNAL cur_addr               : t_tech_ddr_addr; 
  SIGNAL cur_address            : STD_LOGIC_VECTOR(c_address_w-1 DOWNTO 0);
  SIGNAL nxt_cur_address        : STD_LOGIC_VECTOR(c_address_w-1 DOWNTO 0);   
  SIGNAL diff_address           : STD_LOGIC_VECTOR(c_address_w-1 DOWNTO 0);

  SIGNAL addresses_rem          : STD_LOGIC_VECTOR(31 DOWNTO 0); -- nof words (on the user side interface) to rd/wr until end addr is reached
  SIGNAL reg_addresses_rem      : STD_LOGIC_VECTOR(31 DOWNTO 0); -- nof words (on the user side interface) to rd/wr until end addr is reached

  SIGNAL reg_wr_fifo_usedw      : STD_LOGIC_VECTOR(wr_fifo_usedw'RANGE);  -- read side depth of the write FIFO

BEGIN

  dvr_done        <= i_dvr_done; 

  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      state                <= s_init;
      req_burst_cycles     <= (OTHERS => '0');
      i_dvr_done           <= '0';
      cur_address          <= (OTHERS=>'0');
      wr_burst_size        <= 0;
      rd_burst_size        <= 0;
      reg_addresses_rem    <= (OTHERS=>'0');
      reg_wr_fifo_usedw    <= (OTHERS=>'0');
      prev_state           <= s_idle;
    ELSIF rising_edge(clk) THEN
      state                <= nxt_state;
      req_burst_cycles     <= nxt_req_burst_cycles;
      i_dvr_done           <= nxt_dvr_done;
      cur_address          <= nxt_cur_address;
      wr_burst_size        <= nxt_wr_burst_size;
      rd_burst_size        <= nxt_rd_burst_size;
      reg_addresses_rem    <= addresses_rem;
      reg_wr_fifo_usedw    <= wr_fifo_usedw;
      prev_state           <= state;
    END IF;
  END PROCESS;

  -- Add 1 address (accounting for address resulotion) to diff_address: we also want to write the last address. Shift the result right to provide the correct resolution.
  addresses_rem <= RESIZE_UVEC( SHIFT_UVEC( INCR_UVEC(diff_address, g_tech_ddr.rsl), g_tech_ddr.rsl_w), addresses_rem'LENGTH); 
  
  -- End address - current address
  diff_address  <= SUB_UVEC(end_address, cur_address, c_address_w);

  p_burst_size : PROCESS (reg_addresses_rem, reg_wr_fifo_usedw)
    VARIABLE v_burst_size : NATURAL;
  BEGIN 
    -- Write burst size is smallest of g_tech_ddr.maxburstsize, addresses_rem and wr_fifo_usedw
    v_burst_size := g_tech_ddr.maxburstsize+c_margin;
    IF UNSIGNED(reg_wr_fifo_usedw)>=c_margin AND UNSIGNED(reg_addresses_rem)>=c_margin THEN 
      IF v_burst_size > SIGNED('0' & reg_addresses_rem) THEN v_burst_size := TO_UINT(reg_addresses_rem); END IF;
      IF v_burst_size > SIGNED('0' & reg_wr_fifo_usedw) THEN v_burst_size := TO_UINT(reg_wr_fifo_usedw); END IF;
      v_burst_size := v_burst_size - c_margin;
    ELSE
      v_burst_size := 0;
    END IF;
    nxt_wr_burst_size <= v_burst_size;

    -- Read burst size is smallest of g_tech_ddr.maxburstsize and addresses_rem
    v_burst_size := g_tech_ddr.maxburstsize;
    IF UNSIGNED(reg_addresses_rem)>=1 THEN -- prevent assigning <0 value to natural
      IF v_burst_size > SIGNED('0' & reg_addresses_rem) THEN v_burst_size := TO_UINT(reg_addresses_rem); END IF;
    ELSE
      v_burst_size := 0;
    END IF;
    nxt_rd_burst_size <= v_burst_size;
  END PROCESS;

  ctlr_mosi.address <= func_tech_ddr_ctlr_address(cur_addr, g_tech_ddr, c_tech_ddr_ctlr_address_w);
  ctlr_mosi.wrdata <= RESIZE_DDR_CTLR_DATA(wr_snk_in.data);
  ctlr_mosi.burstsize <= ctlr_mosi_burstsize;

  rd_src_out.valid <= ctlr_miso.rdval;
  rd_src_out.data <= RESIZE_DP_DATA(ctlr_miso.rddata);
  
  p_state : PROCESS(prev_state, state, i_dvr_done, ctlr_miso, req_burst_cycles, dvr_wr_not_rd, wr_snk_in, wr_fifo_usedw, wr_burst_size, rd_burst_size, dvr_en, ctlr_init_done, reg_addresses_rem, rd_src_in, ctlr_mosi_burstsize, start_address, cur_address)
  BEGIN  
    nxt_state              <= state;   
    ctlr_mosi.wr           <= '0';
    ctlr_mosi.rd           <= '0';
    ctlr_mosi.burstbegin   <= '0'; 
    ctlr_mosi_burstsize    <= (OTHERS => '0');
    nxt_req_burst_cycles   <= req_burst_cycles;
    wr_snk_out.ready       <= '0';
    nxt_dvr_done           <= i_dvr_done;
    nxt_cur_address        <= cur_address;

    CASE state IS
     
      WHEN s_wr_burst => -- Performs the burst portion (word 2+)        
        ctlr_mosi.wr <= '1';        
        IF ctlr_miso.waitrequest_n = '1' THEN -- when local_ready goes low, that cycle does not count as a burst cycle          
          nxt_req_burst_cycles <= INCR_UVEC(req_burst_cycles, -1);
          wr_snk_out.ready     <= '1'; -- wr side uses latency of 0, so wr_snk_out.ready<='1' acknowledges a successful write request.
          IF UNSIGNED(req_burst_cycles) = 1 THEN -- Then we're in the last cycle of this burst sequence
            nxt_state <= s_wr_request; -- We can only initiate a burst through the wr_request state
          END IF;
        END IF;            

      WHEN s_wr_request => -- Performs 1 write or read and goes into s_wr_burst when requested write words >1         
        nxt_state <= s_wait3; 
        IF UNSIGNED(reg_addresses_rem) = 0 THEN -- end address reached
          nxt_dvr_done  <= '1';              
          nxt_state     <= s_idle;          
        ELSIF ctlr_miso.waitrequest_n = '1' THEN 
          IF wr_snk_in.valid = '1' THEN
            -- Always perform 1st write here             
            wr_snk_out.ready     <= '1';
            ctlr_mosi.wr         <= '1';
            ctlr_mosi.burstbegin <= '1'; -- assert burst begin: strictly this is a burst of 1. 
            ctlr_mosi_burstsize  <= TO_DDR_CTLR_BURSTSIZE(1); -- Set ctlr_mosi_burstsize to 1 by default
            IF wr_burst_size > 1 THEN
              -- Perform any remaining writes in a burst
              nxt_state            <= s_wr_burst;
              nxt_req_burst_cycles <= TO_DDR_CTLR_BURSTSIZE(wr_burst_size-1); -- Forward the required nof burst cycles (-1 as we've done the 1st in this state already) to burst state
              ctlr_mosi_burstsize  <= TO_DDR_CTLR_BURSTSIZE(wr_burst_size  );
            END IF; -- ELSE: there is only 1 word, so no need for remaining burst   
            nxt_cur_address   <= INCR_UVEC(cur_address, UNSIGNED(ctlr_mosi_burstsize)*g_tech_ddr.rsl);
--            IF UNSIGNED(ctlr_mosi_burstsize) = 1 THEN -- Prevents FSM from going into this state again too soon (reg_addresses_rem invalid)
--            nxt_state <= s_wait3; 
--            END IF;
          END IF;
        END IF;        

      WHEN s_rd_request => -- Posts a read request for a burst (0...g_tech_ddr.maxburstsize)      
        nxt_state <= s_wait3;   
        IF UNSIGNED(reg_addresses_rem) = 0 THEN -- end address reached
          nxt_dvr_done  <= '1';              
          nxt_state     <= s_idle;
        ELSE 
          IF rd_src_in.ready = '1' THEN -- Fifo uses its internal almost_full signal to toggle its snk_out.rdy     
            IF ctlr_miso.waitrequest_n = '1' THEN    
              ctlr_mosi.rd         <= '1';                   
              ctlr_mosi.burstbegin <= '1'; -- assert burst begin: strictly this is a burst of 1.                  
              ctlr_mosi_burstsize  <= TO_DDR_CTLR_BURSTSIZE(rd_burst_size);
              IF rd_burst_size = 0 THEN ctlr_mosi_burstsize <= TO_DDR_CTLR_BURSTSIZE(1); END IF;
              nxt_cur_address   <= INCR_UVEC(cur_address, UNSIGNED(ctlr_mosi_burstsize)*g_tech_ddr.rsl);  
--              IF UNSIGNED(ctlr_mosi_burstsize) = 1 THEN -- Prevents FSM from going into this state again too soon (reg_addresses_rem invalid)
--              nxt_state <= s_wait3; 
--              END IF;
            END IF;           
          END IF;
        END IF; 

      -- This wait state is inserted between two requests when necessary, e.g. when FSM enters wr_request
      -- from the state wr_request, an extra cycle is needed for reg_addresses_rem to be valid.
      WHEN s_wait3 =>
        IF prev_state = s_wr_request THEN nxt_state <= s_wr_request; END IF;
        IF prev_state = s_rd_request THEN nxt_state <= s_rd_request; END IF;
   
     -- In this cycle reg_addresses_rem is valid. This cycle is added so wr_burst_size and rd_burst_size
     -- (derived from reg_addresses_rem) are valid the next cycle.
     WHEN s_wait2 =>
       IF dvr_wr_not_rd = '1' THEN
         nxt_state <= s_wr_request; 
       ELSE
         nxt_state <= s_rd_request;             
       END IF;
      
      -- Wait a cycle so reg_addresses_rem is valid the next cyle.
      WHEN s_wait1 =>
        nxt_state <= s_wait2;

      WHEN s_idle =>
        IF dvr_en = '1' THEN  
          nxt_cur_address <= start_address; 
          nxt_dvr_done    <= '0';
          nxt_state       <= s_wait1;
        END IF;
     
      WHEN OTHERS => -- s_init
        IF ctlr_init_done = '1' THEN
          nxt_state <= s_idle;
        END IF;
      
    END CASE;
  END PROCESS;

  start_address <= func_tech_ddr_dq_address(dvr_start_addr, g_tech_ddr, c_address_w);
  end_address   <= func_tech_ddr_dq_address(dvr_end_addr,   g_tech_ddr, c_address_w);
  
  cur_addr <= func_tech_ddr_dq_address(cur_address, g_tech_ddr);

END str;

